/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
package org.greenstone.gatherer.file;

import java.io.*;
import javax.swing.table.*;
import org.greenstone.gatherer.Dictionary;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.gui.FileAssociationDialog;
import org.greenstone.gatherer.msm.MSMUtils;
import org.greenstone.gatherer.util.Utility;
import org.greenstone.gatherer.util.WinRegistry;
import org.w3c.dom.*;

public class FileAssociationManager 
    extends AbstractTableModel {
    static final public String FILENAME_ARG = "%1";
    static final private String DATA_FILENAME = "associations.xml";
    static final private String ENTRY_ELEMENT = "Entry";
    static final private String ESCAPE = "\\\\"; // '\'
    static final private String ESCAPED_ESCAPE = "\\\\\\\\"; // '\\'
    static final private String EXTENSION_ATTRIBUTE = "extension";
    private Document document;
    private File data_file;   

    public FileAssociationManager() {
	// If a associations.xml is available in the GLI install directory load it.
	data_file = new File(Utility.BASE_DIR + DATA_FILENAME);
	if(data_file.exists()) {
	    document = Utility.parse(data_file, true);
	}
	// Load the default associations xml data file. This can be done using the classloader.
	else {
	    document = Utility.parse(Utility.XML_DIRECTORY + DATA_FILENAME, true);
	}
	// Initialize the associations. This involves looking through all current associations searching for those with a command of "".
	if(document != null) {
	    NodeList entries = (document.getDocumentElement()).getElementsByTagName(ENTRY_ELEMENT);
	    for(int i = 0; i < entries.getLength(); i++) {
		Element entry = (Element) entries.item(i);
		String extension = entry.getAttribute(EXTENSION_ATTRIBUTE);
		String command = MSMUtils.getValue(entry);
		// If we encounter a command of "", and we are on windows, we try to automatically set this command.
		if(command.length() == 0 && Utility.isWindows()) {
		    // Create a dummy filename with the appropriate extension
		    String dummy_file = "dummy." + extension;
		    command = WinRegistry.openCommand(dummy_file);
		    // If this succeeded add the association.
		    if(command != null) {
			// Remember to replace the dummy filename with %1
			command = command.replaceAll(dummy_file, FILENAME_ARG);
			// Replace the text in this node.
			MSMUtils.setValue(entry, command);
		    }
		    dummy_file = null;
		}
		command = null;
		extension = null;
		entry = null;
	    }
	    entries = null;
	}
	else {
	    Gatherer.println("Didn't parse anything. About to crash.");
	}
    }
    
    public void edit() {
	FileAssociationDialog dialog = new FileAssociationDialog(this);
	dialog.display(null);
	dialog = null;		  
    }

    public String getBrowserCommand(String url) {
	String command = null;
	if(Utility.isWindows()) {
	    // we use cmd and start
	    if (Utility.isWindows9x()) {
		command = "command.com /c start \"\" \""+url+"\"";
	    } else {
		command = "cmd.exe /c start \"\" \""+url+"\"";
	    }
	    return command;
	}
	// for now uder linux get the fileassoc thing. but eventually want a separate thing.
	// pretend we are tyring to open an html file
	String extension = "html";
	Element entry = getCommand(extension);
	if (entry != null) {
	    command = MSMUtils.getValue(entry); 
	}
	
	if(command == null || command.length() == 0) {
	    FileAssociationDialog dialog = new FileAssociationDialog(this);
	    command = dialog.display(extension);
	    dialog = null;
	}
	if(command != null) {
	    // If no previous entry existed create one.
	    if(entry == null) {
		entry = document.createElement(ENTRY_ELEMENT);
		entry.setAttribute(EXTENSION_ATTRIBUTE, extension);
		document.getDocumentElement().appendChild(entry);
	    }
	    // Replace the text in this node. Remember to replace the dummy filename with %1
	    MSMUtils.setValue(entry, command.replaceAll("temp.html", FILENAME_ARG));
	}
	
	// if we haven't got a command by now, we'll never get one
	if (command == null) return null;
	// Replace %1 with the url in quotes
	command = command.replaceAll(FILENAME_ARG, url);
	return command;
    }

    public int getColumnCount() {
	return 2;
    }

    public String getColumnName(int column) {
	String name;
	switch(column) {
	case 0:
	    name = Dictionary.get("FileAssociationDialog.Table.Extension");
	    break;
	default:
	    name = Dictionary.get("FileAssociationDialog.Table.Command");
	}
	return name;
    }
   
    public String getCommand(File file) {
	String command = null;
	if(file.isFile()) {
	    // Determine extension
	    String filename = file.getAbsolutePath();
	    String extension = filename.substring(filename.lastIndexOf(".") + 1);
	    // Try to retrieve a value from cache
	    Element entry = getCommand(extension);
	    if(entry != null) {
		command = MSMUtils.getValue(entry); 
	    }
	    if(command == null || command.length() == 0) {
		// If command is null, and we are on windows try searching the registry.
		if(Utility.isWindows()) {
		    command = WinRegistry.openCommand(filename);
		}
		// Otherwise display the dialog and ask the user to enter launching command.
		if(command == null || command.length() == 0) {
		    // Show the dialog which forces a user to select the launch command for a certain file.
		    FileAssociationDialog dialog = new FileAssociationDialog(this);
		    command = dialog.display(extension);
		    dialog = null;
		}
		// Hopefully by now we have a command, or else we're never going to get one. Add the association.
		if(command != null) {
		    // If no previous entry existed create one.
		    if(entry == null) {
			entry = document.createElement(ENTRY_ELEMENT);
			entry.setAttribute(EXTENSION_ATTRIBUTE, extension);
			document.getDocumentElement().appendChild(entry);
		    }
		    // Replace the text in this node. Remember to replace the dummy filename with %1
		    MSMUtils.setValue(entry, command.replaceAll(filename, FILENAME_ARG));
		}
	    }
	    if(command != null) {
		// We have to fix filename under windows to escape the backslashes.
		filename = filename.replaceAll(ESCAPE, ESCAPED_ESCAPE);
		// Replace %1 with the appropriate filename
		command = command.replaceAll(FILENAME_ARG, filename);
	    }
	    
	    entry = null;
	    extension = null;
	    filename = null;
	}
	return command;
    }

    public Element getCommand(String target_extension) {
	NodeList entries = document.getDocumentElement().getElementsByTagName(ENTRY_ELEMENT);
	for(int i = 0; i < entries.getLength(); i++) { 
	    Element entry = (Element) entries.item(i);
	    String extension = entry.getAttribute(EXTENSION_ATTRIBUTE);
	    if(extension.equalsIgnoreCase(target_extension)) {
		entries = null;
		extension = null;
		return entry;
	    }
	}
	entries = null;
	return null;
    }

    public String getCommandString(String target_extension) {
	Element entry = getCommand(target_extension);
	if(entry != null) {
	    return MSMUtils.getValue(entry);
	}
	else {
	    return "";
	}
    }

    public String getExtension(int index) {
	NodeList entries = document.getDocumentElement().getElementsByTagName(ENTRY_ELEMENT);
	if(0 <= index && index < entries.getLength()) {
	    Element entry = (Element) entries.item(index);
	    return entry.getAttribute(EXTENSION_ATTRIBUTE);
	}
	return "";
    }

    public int getRowCount() {
	NodeList entries = document.getDocumentElement().getElementsByTagName(ENTRY_ELEMENT);
	return entries.getLength();
    }

    public Object getValueAt(int row, int column) {
	String extension = getExtension(row);
	switch(column) {
	case 0:
	    return extension;
	default:
	    return getCommandString(extension);
	}
    }

    public void save() {
	Utility.export(document, data_file);
    }

    public void setCommand(String extension, String command) {
	Gatherer.println("Set Launch: " + extension + " with " + command);
	// Retrieve any existing entry for this extension
	Element entry = getCommand(extension);
	// If no previous entry existed create one.
	if(entry == null && command != null) {
	    entry = document.createElement(ENTRY_ELEMENT);
	    entry.setAttribute(EXTENSION_ATTRIBUTE, extension);
	    document.getDocumentElement().appendChild(entry);
	}

	if(command != null) {
	    // Replace the text in this node. If the user has used filename instead of %1 then too bad.
	    MSMUtils.setValue(entry, command);
	}
	else {
	    // Remove the entry
	    document.getDocumentElement().removeChild(entry);
	}
	entry = null;
	fireTableDataChanged(); // Can't be anymore efficient as DOM does not gareuntee ordering of new child nodes is consistant
    }

    public int size() {
	NodeList entries = document.getDocumentElement().getElementsByTagName(ENTRY_ELEMENT);
	return entries.getLength();
    }
}
